require 'tubesheet'
require 'chemdriver'
require 'tsutils'
require 'yaml'
require 'sqlite3'
require 'zlib'
require 'chemexam'

include TSUtils

case ARGV.size
when 1
  $param_yaml_file = ARGV[0]

  die "Cannot read yaml file #{$param_yaml_file}, aborting." \
    unless File.exists? $param_yaml_file

  yaml_data = File.open($param_yaml_file) { |f| f.read }

  $yaml_data = yaml_data

else
  die <<-EOF
Usage: #{$0} <yaml_file>
  EOF
end

$driver = Tubesheet::ChemDriver.from_yaml(yaml_data)
die "RNG not initialized!" if $driver.rng.nil?

RNG_SEED = $driver.rng.info[:seed]
TASK     = $driver.rng.info[:task] || RNG_SEED
PREFIX   = $driver.user_config['prefix'] ? ($driver.user_config['prefix'] +"_") : ""

NUM_PROTOFILAMENTS = $driver.num_protofilaments 
MAX_DIMERS         = $driver.max_dimers
INITIAL_OFFSET     = $driver.user_config['initial_offset']
INITIAL_DIMERS     = $driver.user_config['initial_dimers']
INITIAL_STRAIGHT   = $driver.user_config['initial_straight']

MER_GTP_S = Tubesheet::ChemDriver::MONOMER_GTP_STRAIGHT
MER_GDP_S = Tubesheet::ChemDriver::MONOMER_GDP_STRAIGHT

def initialize_lattice
  offset_dimer = MAX_DIMERS - INITIAL_DIMERS
  offset_row = 4*(offset_dimer)
  seam_col = 2*NUM_PROTOFILAMENTS - 1

  for pf in (0...NUM_PROTOFILAMENTS) do
    mer_col = 2*pf
    lat_col = mer_col+1

    for dimer in (0...INITIAL_DIMERS) do
      inter_row = offset_row + 4*dimer
      alpha_row = inter_row+1
      intra_row = alpha_row+1
      beta_row  = intra_row+1

      if dimer < 5
        alpha_mer = :C
        beta_mer  = :c
        long = :C
        lat = 0
      else
        alpha_mer = :S
        beta_mer  = :s
        long = :S
        lat = :t
      end

      if dimer > 0
        $driver[inter_row,mer_col] = long
      end

      $driver[alpha_row,mer_col] = alpha_mer
      $driver[intra_row,mer_col] = long
      $driver[beta_row ,mer_col] = beta_mer

      if lat_col < seam_col
        $driver[alpha_row,lat_col] = lat
        $driver[beta_row,lat_col] = lat
      end
    end
  end

  for dimer in (0...INITIAL_DIMERS-1) do
    beta_row   = offset_row + 4*dimer + 3
    alpha_next = beta_row+2

    (r0,c0) = $driver.wrap_coords(beta_row  , seam_col+1)
    (r1,c1) = $driver.wrap_coords(alpha_next, seam_col+1)

    if (r0 >= 1) && (r1+1 < $driver.nrow)
      if is_straight?($driver[r0,c0]) && is_straight?($driver[r1,c1])
        $driver[beta_row  ,lat_col] = :t
        $driver[alpha_next,lat_col] = :t
      end
    end

  end
end

$last_accepts = 0
$last_calls = 0

def print_status(f,driver,energy,avg_plus,var_plus,avg_minus,var_minus)
  if $fast_sheet_model
    accepts   = $fast_sheet_model.num_hook_accepts - $last_accepts
    ncalls    = $fast_sheet_model.num_hook_calls   - $last_calls
    tube      = $fast_sheet_model.tube_last_rows
    mech_ener = $fast_sheet_model.energy
    (avgtube,_,sdtube) = calc_mean_sd(tube)

    $last_accepts = $fast_sheet_model.num_hook_accepts
    $last_calls   = $fast_sheet_model.num_hook_calls
  else
    accepts   = 0
    ncalls    = 0
    mech_ener = 0.0
    avgtube   = 0.0
    sdtube   = 0.0
  end

  npf = driver.num_protofilaments

  min_tube = Array.new(npf, 0.0)
  max_tube = Array.new(npf, 0.0)
  row_range = driver.find_row_range
  (0...npf).each do |pf|
    col = 2*pf
    (min_row,max_row) = row_range[col]
    min_tube[pf] = max_row / 4
    for alpha in min_row.step(max_row,4) do
      beta = alpha+2
      if is_straight?($driver[alpha,col]) && is_straight?($driver[beta,col])
        min_tube[pf] = (alpha+3)/4
        break
      end
    end

    max_tube[pf] = min_row / 4
    for beta in max_row.step(min_row,-4) do
      alpha = beta - 2
      if is_straight?($driver[alpha,col]) && is_straight?($driver[beta,col])
        max_tube[pf] = (alpha+3)/4
        break
      end
    end
  end
  (avg_mintube,_,sd_mintube) = calc_mean_sd(min_tube)
  (avg_maxtube,_,sd_maxtube) = calc_mean_sd(max_tube)

  f.puts 'Time %10.2f  Steps %8d  Energy %7.2f  E(fsm) %7.2f  PlusTip %8.2f +/- %5.2f MinusTip %8.2f +/- %5.2f  Tube %8.2f ± %5.2f  ShAcc %6d/%6d (%7.2f) avg_dt %12.5f' % [
    driver.time, driver.nsteps, energy, mech_ener,
    avg_plus, Math::sqrt(var_plus),
    avg_minus, Math::sqrt(var_minus),
    avg_mintube,sd_mintube,
    accepts,ncalls,100.0*(accepts.to_f)/ncalls.to_f,
    1.0/driver.k_total
  ]
end

def write_output(driver, status_f, state_f, always_write=false)
  en = $driver.energy(:calcpops => false)
  if en.nan?
    puts "ERROR IN SIMULATION!  Energy is NaN! Aborting!"
    exit 1
  end
  (reached_end, avg_plus, var_plus, avg_minus, var_minus) = calc_tip_location($driver)

  if always_write || !reached_end
    print_status(status_f, $driver, en,avg_plus, var_plus, avg_minus, var_minus)
    print_status(STDOUT,$driver, en, avg_plus, var_plus, avg_minus, var_minus)
    STDOUT.flush
    $driver.write_state(state_f)
  end

  return reached_end
end

STATUS_FMT = '%slattice-%d.txt'
STATE_FMT  = '%slattice-%d.lst'

status_filename  = STATUS_FMT % [PREFIX,TASK]
state_filename   = STATE_FMT  % [PREFIX,TASK]

def calc_mean_sd(lst)
  avg = lst.reduce(&:+).to_f/lst.size
  if lst.size > 1
    var  = lst.map{|x| (x - avg)**2}.reduce(&:+) / (lst.size-1)
    sd = Math::sqrt(var)
  else
    var = -1.0
    sd = -1.0
  end
  [avg,var,sd]
end

def calc_tip_location(driver)
  npf = driver.num_protofilaments
  ndimers = driver.num_dimers
  tiplocs = driver.find_row_range.
    values_at(*(0.step(driver.ncol-1,2)))

  minus_ends = tiplocs.map { |item| (3+item[0])/4 }
  plus_ends  = tiplocs.map { |item| (1+item[1])/4 }

  (avg_plus ,var_plus ,sd_plus ) = calc_mean_sd(plus_ends.reject {|x| x < 0 } )
  (avg_minus,var_minus,sd_minus) = calc_mean_sd(minus_ends.reject {|x| x < 0 })

  dist = (0 ... minus_ends.size).map { |ind| plus_ends[ind] - minus_ends[ind] }
  reached_minus = minus_ends.any? { |loc| loc <= 1 }
  reached_plus  = false
  too_short = dist.any? { |l| l < 2 } 
  reached_end = reached_minus || reached_plus || too_short

  [reached_end,avg_plus,var_plus,avg_minus,var_minus]
end

def make_model(driver, strain_scale=1.0)
  npf = NUM_PROTOFILAMENTS

  config = driver.user_config['fast_sheet_model']

  is_enabled = config['enabled']
  return nil unless is_enabled && is_enabled != false && is_enabled != 0
  scale = config['scale'] ? Float(config['scale']) : 1.0

  [ [ 'ifc_model',          6 ],
    [ 'existing_model',     3 ],
    [ 'relax_coeff',      npf ],
    ['no_tube_strain',    npf ],
    ['left_tube_strain',  npf ],
    ['right_tube_strain', npf ],
    ['both_tube_strain',  npf ]
  ].each do |sec, sz|
    die "Missing 'fast_sheet_model' subsection: '#{sec}'" \
      unless config[sec]

    okay = (Array === config[sec]) \
      && (config[sec].size == sz)  \
      && (config[sec].map { |x| (Numeric === x) }.all?)

    die "'fast_sheet_model' subsection '#{sec}' requires #{sz} numbers" unless okay
  end

  ifc_model         = config['ifc_model'        ]
  existing_model    = config['existing_model'   ]
  relax_coeff       = config['relax_coeff'      ]

  no_tube_strain    = config['no_tube_strain'   ].map { |x| x*strain_scale }
  left_tube_strain  = config['left_tube_strain' ].map { |x| x*strain_scale }
  right_tube_strain = config['right_tube_strain'].map { |x| x*strain_scale }
  both_tube_strain  = config['both_tube_strain' ].map { |x| x*strain_scale }

  model = Tubesheet::FastSheetModel.new(
    driver, 
    :ifc_model      => ifc_model,
    :existing_model => existing_model,
    :relax_coeff    => relax_coeff,
    :no_tube_strain => no_tube_strain,
    :left_tube_strain => left_tube_strain,
    :right_tube_strain => right_tube_strain,
    :both_tube_strain => both_tube_strain
  )

  fluct_dist = config['sheet_fluct_dist']
  fluct_penalty = config['sheet_fluct_penalty']
  model.sheet_fluctuation_dist = fluct_dist if fluct_dist
  model.sheet_fluctuation_penalty = fluct_penalty if fluct_penalty

  model.scale = scale
  model
end

SIM_START_TIME = Time.now
begin
  status_f = nil
  state_f  = nil

  status_f = File.open(status_filename, 'w')
  state_f  = File.open(state_filename , 'w')

  status_f.puts 'seed %d' % [RNG_SEED]
  puts 'seed is %d' % [RNG_SEED]

  initialize_lattice
  $driver.rebuild_moves!

  $fast_sheet_model = make_model($driver)

  $driver.hydrolysis_only_at_sss = true

  if $fast_sheet_model
    puts "Enabling fast sheet model"
    $fast_sheet_model.add_movehook!
  else
    puts "Fast sheet model disabled"
  end

  iter = 0
  t_next = 0.0
  t_max = $driver.t_max
  dt = $driver.dt
  while $driver.time < t_max do
    reached_end = write_output($driver,status_f,state_f)
    break if reached_end

    iter += 1

    if iter % 10 == 0
      status_f.flush
      state_f.flush
    end

    GC.start

    if dt <= 0.0
      $driver.single_step
    else
      avg_dt = 1.0/$driver.k_total
      twice_avg_dt = 2*avg_dt
      eff_dt = (dt >= twice_avg_dt) ? dt : twice_avg_dt
      while t_next <= $driver.time
        t_next += eff_dt
      end
      $driver.simulate_until(t_next)  
    end
  end

  write_output($driver,status_f,state_f, true)
rescue Exception => e
  puts "Error during simuation: #{e}"
  puts "  "+e.backtrace.join("\n  ")
ensure
  SIM_END_TIME = Time.now
  status_f.close if status_f
  state_f.close  if state_f
end

time_secs = SIM_END_TIME - SIM_START_TIME
mins = (time_secs/60.0).floor
secs = time_secs - (60.0 * mins)
hours = (mins/60.0).floor
mins -= hours*60.0

puts "Simulation took #{time_secs} seconds (#{hours} hours, #{mins} minutes, #{secs} seconds)"
puts "DONE"

